"""
****************************************************************************************************
:copyright (c) 2019-2021 URBANopt, Alliance for Sustainable Energy, LLC, and other contributors.
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:
Redistributions of source code must retain the above copyright notice, this list of conditions
and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the
distribution.
Neither the name of the copyright holder nor the names of its contributors may be used to endorse
or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
****************************************************************************************************
"""


# ditto_reader_cli

import click
import json
import sys
from pathlib import Path
from urbanopt_ditto_reader.urbanopt_ditto_reader import UrbanoptDittoReader

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    """URBANopt Ditto Reader"""
    pass

@cli.command(short_help="Run OpenDSS on existing urbanopt simulations")
@click.option(
    '-s',
    '--scenario_file',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Path to scenario file"
)
@click.option(
    '-f',
    "--feature_file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Path to feature file"
)
@click.option(
    "-e",
    "--equipment",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Path to optional custom equipment file"
)
@click.option(
    "-b",
    "--start_time",
    type=str,
    default=None,
    help="Beginning timestamp of simulation. If invalid or None, all timepoints will be run"
)
@click.option(
    "-n",
    "--end_time",
    type=str,
    default=None,
    help="Ending timestamp of simulation. If invalid or None, all timepoints will be run"
)
@click.option(
    "-t",
    "--timestep",
    type=float,
    default=None,
    help="Interval between simulations in minutes."
)

@click.option(
    '-r',
    '--reopt',
    is_flag=True,
    help="Flag to use REopt data in this openDSS analysis"
)
@click.option(
    '-c',
    '--config',
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Path to a json config file for all settings"
)
@click.option(
    '-u',
    '--upgrade',
    is_flag=True,
    help="Flag to automatically upgrade transformers if undersized"
)

def run_opendss(scenario_file, feature_file, equipment, start_time, end_time, timestep, reopt, config, upgrade):
    """
    \b
    Run OpenDSS on an existing URBANopt scenario.
    If referencing your own json config file: all settings must be made in that file using absolute paths.

    \f
    :param scenario_file: Path, location and name of scenario csv file
    :param feature_file: Path, location & name of feature json file
    :param equipment: Path, location and name of custom equipment file
    :param start_time: String, timestamp of the start time of the simulation. Uses format "YYYY/MM/DD HH:MM:SS". Cross referenced with the timestamps in the SCENARIO_NAME/opendss/profiles/timestamps.csv file created from profiles in SCENARIO_NAME/FEATURE_ID/feature_reports/feature_report_reopt.csv if use_reopt is true and SCENARIO_NAME/FEATURE_ID/feature_reports/default_feature_report.csv if use_reopt is false. It runs the entire year if timestep not found.
    :param end_time: String, timestamp of the end time of the simulation. Uses format "YYYY/MM/DD HH:MM:SS". Cross referenced with the timestamps in the SCENARIO_NAME/opendss/profiles/timestamps.csv file created from profiles in SCENARIO_NAME/FEATURE_ID/feature_reports/feature_report_reopt.csv if use_reopt is true and SCENARIO_NAME/FEATURE_ID/feature_reports/default_feature_report.csv if use_reopt is false. It runs the entire year if timestep not found.
    :param timestep: Float, number of minutes between each simulation. If larger than timesteps provided by the reopt feature reports (if use_repot is true), or urbanopt feature reports (if use_reopt is false), an error is raised
    :param reopt: Boolean, flag to specify that reopt data is present and OpenDSS analysis should include it
    :param upgrade: Boolean, flag to automatically ugrade transformers that are smaller than the sum of the peak loads that they serve.
    :param config: Path, location of config file specifying input options for OpenDSS
    """

    try:
        if config:
            with open(config) as f:
                config_dict = json.load(f)
        else:
            scenario_name = Path(scenario_file).stem
            scenario_dir = Path(scenario_file).parent / "run" / scenario_name

            config_dict = {
                'urbanopt_scenario_file': scenario_file,
                'urbanopt_geojson_file': feature_file,
                'use_reopt': reopt,
                'opendss_folder': scenario_dir / 'opendss',
                'start_time': start_time,
                'end_time': end_time,
                'timestep': timestep,
                'upgrade_transformers': upgrade
                }
            if equipment:
                config_dict['equipment_file'] = equipment

        ditto = UrbanoptDittoReader(config_dict)
        ditto.run()
    except Exception as e:
        print(f"CLI failed with message: {e}")
        sys.exit(1)
    else:
        print(f"\nDone. Results located in {config_dict['opendss_folder']}\n")



"""
****************************************************************************************************
:copyright (c) 2019-2021 URBANopt, Alliance for Sustainable Energy, LLC, and other contributors.
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:
Redistributions of source code must retain the above copyright notice, this list of conditions
and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the
distribution.
Neither the name of the copyright holder nor the names of its contributors may be used to endorse
or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
****************************************************************************************************
"""

import datetime
import math
import logging
import os
import pandas as pd
import json
import networkx as nx

from ditto.readers.abstract_reader import AbstractReader
from ditto.store import Store
from ditto.models.node import Node
from ditto.models.line import Line
from ditto.models.wire import Wire
from ditto.models.powertransformer import PowerTransformer
from ditto.models.winding import Winding
from ditto.models.phase_winding import PhaseWinding
from ditto.models.base import Unicode
from ditto.models.position import Position
from ditto.models.feeder_metadata import Feeder_metadata
from ditto.models.capacitor import Capacitor
from ditto.models.phase_capacitor import PhaseCapacitor
from ditto.models.load import Load
from ditto.models.phase_load import PhaseLoad
from ditto.models.power_source import PowerSource
from ditto.network.network import Network
from ditto.models.photovoltaic import Photovoltaic
from ditto.models.timeseries import Timeseries
from ditto.models.feeder_metadata import Feeder_metadata
from ditto.modify.modify import Modifier

class Reader(AbstractReader):
    """
    Reader for the Urbanopt geojson file with supporting database files
    """
    register_names = ["geojson","GeoJson"]

    def __init__(self, **kwargs):
        super(Reader,self).__init__(**kwargs)

        if "geojson_file" in kwargs:
            self.geojson_file = kwargs["geojson_file"]
            self.geojson_content = None
        else:
            raise ValueError("No geojson_file parameter provided")
        if "equipment_file" in kwargs:
            self.equipment_file = kwargs["equipment_file"]
            self.equipment_data = None
        else:
            raise ValueError("No equipment_file parameter provided")
        if "load_folder" in kwargs:
            self.load_folder = kwargs["load_folder"]
            self.load_data = None
        else:
            raise ValueError("No load_folder parameter provided")
        if "use_reopt" in kwargs:
            self.use_reopt = kwargs["use_reopt"]
        else:
            self.use_reopt = False
            print("Warning - using default urbanopt configuration")

        self.is_timeseries = False
        self.timeseries_location = None
        self.relative_timeseries_location = None
        if 'is_timeseries' in kwargs:
            self.is_timeseries = kwargs['is_timeseries']
            if 'timeseries_location' in kwargs:
                self.timeseries_location = kwargs['timeseries_location']
            if 'relative_timeseries_location' in kwargs:
                self.relative_timeseries_location = kwargs['relative_timeseries_location']



    def get_geojson_data(self, filename):
        """
        Helper method to save all the json data in the geojson file
        """
        content = []
        try:
            with open(filename,"r") as f:
                content = json.load(f)
        except:
            raise IOError("Problem trying to read json from file "+filename)
        return content

    def get_equipment_data(self, filename):
        """
        Helper method to save all the json data in the equipment file
        """
        content = []
        try:
            with open(filename,"r") as f:
                content = json.load(f)
        except:
            raise IOError("Problem trying to read json from file "+filename)
        return content

    def get_feature_data(self, filename):
        """
        Helper method to save all the json data in the equipment file
        """
        content = []
        try:
            with open(filename,"r") as f:
                content = json.load(f)
        except:
            raise IOError("Problem trying to read json from file "+filename)
        return content

    def parse(self, model, **kwargs):
        """General parse function.
        Responsible for calling the sub-parsers and logging progress.
        :param model: DiTTo model
        :type model: DiTTo model
        :param verbose: Set verbose mode. Optional. Default=False
        :type verbose: bool
        :returns: 1 for success, -1 for failure
        :rtype: int
        """

        self.geojson_content = self.get_geojson_data(self.geojson_file)
        self.equipment_data = self.get_equipment_data(self.equipment_file)

        # Call parse from abstract reader class
        try:
            super(Reader, self).parse(model, **kwargs)
        except KeyError:
            raise SystemExit("\nFeatureFile is missing components. Are all electrical features included?")
        return 1

    def parse_lines(self, model, **kwargs):
        """Line parser.
        :param model: DiTTo model
        :type model: DiTTo model
        :returns: 1 for success, -1 for failure
        :rtype: int
        """
        bad_lines = []
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and element['properties']['type'] == 'ElectricalConnector':
                line = Line(model)
                line.name = element['properties']['id']
                if element['properties']['startJunctionId'] in self.substations:
                    line.from_element = 'source'
                else:
                    line.from_element = element['properties']['startJunctionId']
                if element['properties']['endJunctionId'] in self.substations:
                    line.to_element = 'source'
                else:
                    line.to_element = element['properties']['endJunctionId']
                line.length = element['properties']['total_length']*0.3048 #length from feet to meters
                all_wires = []
                if not 'wires' in element['properties'] or len(element['properties']['wires']) == 0:
                    bad_lines.append(line.name)
                    continue

                for wire_type in element['properties']['wires']:
                    found_wire = False
                    for db_wire in self.equipment_data['wires']:
                        if db_wire['nameclass'] == wire_type:
                            found_wire = True
                            wire = Wire(model)
                            wire.nameclass = wire_type.replace(' ','_').replace('/','-')
                            if 'OH' in wire_type:
                                line.line_type = 'overhead'
                            else:
                                line.line_type = 'underground'
                            if 'S1' in wire_type:
                                wire.phase = 'A'
                            elif 'S2' in wire_type:
                                wire.phase = 'B'
                            else:
                                wire.phase = wire_type.split(' ')[-1] #Currently the convention is that the last element is the phase.
                            wire.ampacity = float(db_wire['ampacity'])
                            wire.gmr = float(db_wire['gmr'])*0.3048
                            wire.resistance = float(db_wire['resistance'])*0.3048
                            wire.diameter = float(db_wire['diameter'])*0.3048
                            wire.X = float(db_wire['x'])*0.3048
                            wire.Y = float(db_wire['height'])*0.3048
                            all_wires.append(wire)
                    if not found_wire:
                        raise ValueError(f'No wire found in catalog for {wire_type}')
                line.wires = all_wires


        if len(bad_lines) > 0:
            print('Following lines are missing wires:')
            for line in bad_lines:
                print(line)
            print()
            raise ValueError("Wires missing for some lines")

        return 1

    def parse_nodes(self, model, **kwargs):
        """Node parser.
        :param model: DiTTo model
        :type model: DiTTo model
        :returns: 1 for success, -1 for failure
        :rtype: int
        """
        # Assume one substation per feeder with a single junction
        substation_map = {}
        self.substations = set()

        for element in self.geojson_content["features"]:
            if 'properties' in element and 'DSId' in element['properties'] and 'id' in element['properties']:
                if element['properties']['DSId'] in substation_map:
                    substation_map[element['properties']['DSId']].append(element['properties']['id'])
                else:
                    substation_map[element['properties']['DSId']] = [element['properties']['id']]

        for element in self.geojson_content["features"]:
            if 'properties' in element and 'district_system_type' in element['properties'] and element['properties']['district_system_type'] == 'Electrical Substation':
                if element['properties']['id'] in substation_map:
                    for i in substation_map[element['properties']['id']]:
                        self.substations.add(i)

        if len(self.substations)>1:
            print('Warning - multiple power sources have been added')
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and element['properties']['type'] == 'ElectricalJunction':
                node = Node(model)
                node.name = element['properties']['id']
                if node.name in self.substations:
                    node.nominal_voltage = 13200 #placeholder voltage
                    node.is_substation_connection = True
                    node.setpoint = 1.0
                    node.name = 'source'
                    meta = Feeder_metadata(model)
                    meta.headnode = 'source'
                    meta.nominal_voltage = 13200 #placeholder voltage
                    meta.name = 'urbanopt-feeder'
                    powersource = PowerSource(model)
                    powersource.is_sourcebus = True
                    powersource.name = 'ps_source'
                    powersource.nominal_voltage = 13200
                    powersource.connecting_element = 'source'
                    powersource.per_unit = 1.0
                position = Position(model)
                position.lat = float(element['geometry']['coordinates'][1])
                position.long = float(element['geometry']['coordinates'][0])
                node.positions = [position]


        return 1

    def parse_transformers(self, model, **kwargs):
        """Transformer parser.
        :param model: DiTTo model
        :type model: DiTTo model
        :returns: 1 for success, -1 for failure
        :rtype: int
        """

        # Assume that each transformer has one from node and one to node.

        connection_map = {'Delta':'D','Wye':'Y'}
        transformer_panel_map = {}
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'DSId' in element['properties'] and 'id' in element['properties']:
                if element['properties']['DSId'] in transformer_panel_map:
                    transformer_panel_map[element['properties']['DSId']].append(element['properties']['id'])
                else:
                    transformer_panel_map[element['properties']['DSId']] = [element['properties']['id']]

        source_voltages = set() 
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'district_system_type' in element['properties'] and element['properties']['district_system_type'] == 'Transformer':
                transformer_id = element['properties']['id']
                transformer = PowerTransformer(model)
                if transformer_id in transformer_panel_map:
                    if len(transformer_panel_map[transformer_id]) <2:
                        print(f"No from and to elements found for transformer {transformer_id}")
                    if len(transformer_panel_map[transformer_id]) >2:
                        print("Warning - the transformer "+transformer_id+" should have a from and to element - "+str(len(transformer_panel_map[transformer_id]))+" junctions on the transformer")
                    if len(transformer_panel_map[transformer_id]) >=2:
                        found_transfomer = False
                        for db_transformer in self.equipment_data['transformer_properties']:
                            if element['properties']['equipment'][0] == db_transformer['nameclass']:
                                found_transfomer = True
                                transformer.from_element = transformer_panel_map[transformer_id][0]
                                transformer.to_element = transformer_panel_map[transformer_id][1] #NOTE: Need to figure out correct from and to directions here.
                                transformer.name = transformer_id
                                transformer.reactances = [float(db_transformer['reactance'])]
                                transformer.is_center_tap = db_transformer['is_center_tap']
                                windings = [Winding(model),Winding(model)]
                                connections = db_transformer['connection'].split('-')

                                if transformer.is_center_tap:
                                    windings.append(Winding(model))
                                    transformer.reactances.append(float(db_transformer['reactance']))
                                    transformer.reactances.append(float(db_transformer['reactance'])) #TODO: map reactance values correctly for center-taps
                                for i in range(len(windings)):
                                    phase_windings = []
                                    if transformer.is_center_tap and i >0:
                                        for phase in ['A','B']:
                                            pw = PhaseWinding(model)
                                            pw.phase = phase
                                            phase_windings.append(pw)
                                    else:
                                        for phase in db_transformer['phases']:
                                            pw = PhaseWinding(model)
                                            pw.phase = phase
                                            phase_windings.append(pw)
                                    windings[i].phase_windings = phase_windings
                                    windings[i].rated_power = float(db_transformer['kva'])*1000
                                    if i<1:
                                        windings[i].nominal_voltage = float(db_transformer['high_voltage'])*1000
                                        source_voltages.add(windings[i].nominal_voltage)
                                        if transformer.is_center_tap:
                                            windings[i].nominal_voltage = windings[i].nominal_voltage/(3**0.5)
                                        windings[i].connection_type = connection_map[connections[0]]
                                        windings[i].voltage_type = 0
                                        windings[i].resistance = float(db_transformer['resistance'])
                                    else:
                                        windings[i].nominal_voltage = float(db_transformer['low_voltage'])*1000
                                        windings[i].connection_type = connection_map[connections[1]]
                                        windings[i].voltage_type = 1
                                        windings[i].resistance = float(db_transformer['resistance'])
                                transformer.windings = windings
                        if not found_transfomer:
                            raise ValueError(f'No transfomer found in catalog for {element["properties"]["equipment"][0]}')


        # Note that the source voltage is set to be the highest side of a transformer that is used
        if len(source_voltages) == 1:
            source_voltage = source_voltages.pop()
            model.set_names()
            model['source'].nominal_voltage = source_voltage
            model['ps_source'].nominal_voltage = source_voltage
            model['urbanopt-feeder'].nominal_voltage = source_voltage

        else:
            raise ValueError('Problem setting source voltage. No high transformer values or non-unique high side voltages. Using defaul of 13.2kV')
        return 1

    def parse_capacitors(self, model, **kwargs):
        """Capacitor parser.
        :param model: DiTTo model
        :type model: DiTTo model
        :returns: 1 for success, -1 for failure
        :rtype: int
        """

        return 1

    def parse_loads(self, model, **kwargs):
        """Load parser.
        :param model: DiTTo model
        :type model: DiTTo model
        :returns: 1 for success, -1 for failure
        :rtype: int
        """
        model.set_names()
        network = Network()
        network.build(model,source="source")

        building_map = {}
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and 'buildingId' in element['properties'] and element['properties']['type'] == 'ElectricalJunction':
                building_map[element['properties']['buildingId']] = element['properties']['id']
        disconnected_loads = []
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and element['properties']['type'] == 'Building':
                id_value = element['properties']['id']
                if not id_value in building_map:
                    print(f'Warning - {id_value} missing from building object. Skipping...',flush=True)
                    continue
                connecting_element = building_map[id_value]
                load = Load(model)
                load.name = id_value
                load.connecting_element = connecting_element
                upstream_transformer_name = None
                try:
                    upstream_transformer_name = network.get_upstream_transformer(model,connecting_element)
                except:
                    disconnected_loads.append(element['properties']['id']) #Caused by elements not being connected
                if upstream_transformer_name is not None:
                    upstream_transformer = model[upstream_transformer_name]
                    is_center_tap = upstream_transformer.is_center_tap
                    load.nominal_voltage = upstream_transformer.windings[1].nominal_voltage
                else:
                    print(f'Warning - Load {load.name} is incorrectly connected',flush=True)

                load_path = os.path.join(self.load_folder,id_value,'feature_reports')
                load_multiplier = 1000
                if os.path.exists(load_path): #We've found the load data

                    load_data = None
                    load_column = None
                    if self.use_reopt:
                        load_data = pd.read_csv(os.path.join(load_path,'feature_report_reopt.csv'),header=0)
                        load_column = 'REopt:Electricity:Load:Total(kw)'
                    else:
                        load_data = pd.read_csv(os.path.join(load_path,'default_feature_report.csv'),header=0)
                        load_column = 'Net Power(kW)'
                        if not load_column in load_data:
                            load_column = 'Net Power(W)'
                            load_multiplier = 1
                        if not load_column in load_data:
                            raise ValueError("Neither of the columns 'Net Power(W)' or 'Net Power(kW)' found in default_feature_report.csv'")
                    max_load = max(load_data[load_column])

                    phases = []
                    if upstream_transformer_name is not None:
                        for ph_wdg in upstream_transformer.windings[1].phase_windings:
                            phases.append(ph_wdg.phase)
                        if is_center_tap:
                            phases = ['A','B']

                    phase_loads = []
                    for phase in phases:
                        phase_load = PhaseLoad(model)
                        phase_load.phase = phase
                        power_factor = 0.95
                        phase_load.p = max_load/len(phases) * load_multiplier
                        phase_load.q = phase_load.p * ((1/power_factor-1)**0.5) 
                        phase_loads.append(phase_load)
                    load.phase_loads = phase_loads
                    if self.is_timeseries:
                        data = load_data[load_column]
                        timestamps = load_data['Datetime']
                        delta = datetime.datetime.strptime(timestamps.loc[1],'%Y/%m/%d %H:%M:%S') - datetime.datetime.strptime(timestamps.loc[0],'%Y/%m/%d %H:%M:%S')
                        data_pu = data/max_load
                        if not self.timeseries_location is None:
                            if not os.path.exists(self.timeseries_location):
                                os.makedirs(self.timeseries_location)
                            data.to_csv(os.path.join(self.timeseries_location,'load_'+id_value+'.csv'),header=False, index=False)
                            data_pu.to_csv(os.path.join(self.timeseries_location,'load_'+id_value+'_pu.csv'),header=False, index=False)
                            timestamps.to_csv(os.path.join(self.timeseries_location,'timestamps.csv'),header=True,index=False)
                            timeseries = Timeseries(model)
                            timeseries.feeder_name = load.feeder_name
                            timeseries.substation_name = load.substation_name
                            timeseries.interval = delta.seconds/3600.0 #assume 15 minute loads
                            timeseries.data_type = 'float'
                            timeseries.data_location = os.path.join(self.relative_timeseries_location,'load_'+id_value+'_pu.csv')
                            timeseries.data_label = 'feature_'+id_value
                            timeseries.scale_factor = 1
                            load.timeseries = [timeseries]





                else:
                    print('Load information missing for '+id_value,flush=True)


        if len(disconnected_loads) > 1:
            all_disconnected = ','.join(disconnected_loads)
            #raise ValueError('The following loads have connection problems: '+all_disconnected)
            print('The following loads have connection problems: '+all_disconnected)
        return 1


    def parse_dg(self, model, **kwargs):
        """PV parser.
        :param model: DiTTo model
        :type model: DiTTo model
        :returns: 1 for success, -1 for failure
        :rtype: int
        """
        if not self.use_reopt:
            return 1
        model.set_names()
        network = Network()
        network.build(model,source="source")
        building_map = {}
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and 'buildingId' in element['properties'] and element['properties']['type'] == 'ElectricalJunction':
                building_map[element['properties']['buildingId']] = element['properties']['id']
        for element in self.geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and element['properties']['type'] == 'Building':
                id_value = element['properties']['id']
                if not id_value in building_map:
                    print(f'Warning - {id_value} missing from building object. Skipping...',flush=True)
                    continue
                connecting_element = building_map[id_value]
                try:
                    feature_data = self.get_feature_data(os.path.join(self.load_folder,id_value,'feature_reports','feature_report_reopt.json'))
                except Exception as e:
                    print(e)
                    continue
                pv_kw = feature_data['distributed_generation']['total_solar_pv_kw']

                upstream_transformer_name = network.get_upstream_transformer(model,connecting_element)
                if upstream_transformer_name is not None:
                    upstream_transformer = model[upstream_transformer_name]
                    is_center_tap = upstream_transformer.is_center_tap
                else:
                    print(f'Warning - DG {pv.name} is incorrectly connected',flush=True)
                if pv_kw >0:
                    pv = Photovoltaic(model)
                    pv.name = 'solar_'+id_value
                    pv.connecting_element = connecting_element
                    if upstream_transformer_name is not None:
                        pv.nominal_voltage = upstream_transformer.windings[1].nominal_voltage
                        pv.connection_type = upstream_transformer.windings[1].connection_type
                        phases = []
                        for ph_wdg in upstream_transformer.windings[1].phase_windings:
                            phases.append(Unicode(ph_wdg.phase))
                        if is_center_tap:
                            phases = [Unicode('A'),Unicode('B')]
                        pv.phases = phases
                    pv.rated_power = pv_kw
                    pv.active_rating = 1.1*pv_kw # Should make this a parameter instead
                    if self.is_timeseries:
                        load_data = pd.read_csv(os.path.join(self.load_folder,id_value,'feature_reports','feature_report_reopt.csv'),header=0)
                        data = load_data['REopt:ElectricityProduced:PV:Total(kw)']
                        timestamps = load_data['Datetime']
                        delta = datetime.datetime.strptime(timestamps.loc[1],'%Y/%m/%d %H:%M:%S') - datetime.datetime.strptime(timestamps.loc[0],'%Y/%m/%d %H:%M:%S')
                        data_pu = data/pv_kw
                        if not self.timeseries_location is None:
                            if not os.path.exists(self.timeseries_location): #Should have already been created for the loads
                                os.makedirs(self.timeseries_location)
                            data.to_csv(os.path.join(self.timeseries_location,'pv_'+id_value+'.csv'),header=False, index=False)
                            data_pu.to_csv(os.path.join(self.timeseries_location,'pv_'+id_value+'_pu.csv'),header=False, index=False)
                            timeseries = Timeseries(model)
                            timeseries.feeder_name = pv.feeder_name
                            timeseries.substation_name = pv.substation_name
                            timeseries.interval = delta.seconds/3600.0
                            timeseries.data_type = 'float'
                            timeseries.data_location = os.path.join(self.relative_timeseries_location,'pv_'+id_value+'_pu.csv')
                            timeseries.data_label = 'feature_'+id_value
                            timeseries.scale_factor = 1
                            pv.timeseries = [timeseries]



        return 1



"""
****************************************************************************************************
:copyright (c) 2019-2021 URBANopt, Alliance for Sustainable Energy, LLC, and other contributors.
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:
Redistributions of source code must retain the above copyright notice, this list of conditions
and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the
distribution.
Neither the name of the copyright holder nor the names of its contributors may be used to endorse
or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
****************************************************************************************************
"""

import datetime
import json
import math
import os
from pathlib import Path
import pandas as pd
import opendssdirect as dss


class UrbanoptDittoReader(object):
    def __init__(self, config_data={}):

        self.module_path = Path(__file__).parent.parent

        # load default config from config.json
        default_data = self.default_config()

        # merge with whatever came in on config_data
        config = {**default_data, **config_data}

        print("CONFIGS used: {}".format(config))

        self.geojson_file = str(Path(config['urbanopt_geojson_file']).expanduser().resolve())
        self.urbanopt_scenario_name = Path(config['urbanopt_scenario_file']).stem
        self.urbanopt_scenario = str(Path(config['urbanopt_scenario_file']).expanduser().parent / 'run' / self.urbanopt_scenario_name)
        self.equipment_file = str(Path(config['equipment_file']).expanduser().resolve())
        self.dss_analysis = str(Path(config['opendss_folder']).expanduser().resolve())
        self.use_reopt = config['use_reopt']
        self.start_time = None
        self.end_time = None
        self.timestep = None
        if 'timestep' in config:
            self.timestep = config['timestep']
        if 'start_time' in config and 'end_time' in config:
            self.start_time = str(config['start_time'])
            self.end_time = str(config['end_time'])
        else:
            print('No start_time and end_time provided. Running all timepoints')

        self.upgrade_transformers = False
        if 'upgrade_transformers' in config:
            self.upgrade_transformers = config['upgrade_transformers']


        self.timeseries_location = os.path.join(self.dss_analysis, 'profiles')

    def default_config(self):
        """Read default config file"""
        example_config_file = Path(__file__).parent / 'example_config.json'
        with open(example_config_file) as f:
            default_data = self.fix_paths(json.load(f))
        return default_data

    def fix_paths(self, data):
        """Fix data to be relative path wrt this module"""
        for k, v in data.items():
            if k == 'use_reopt' or k == 'start_time' or k == 'end_time' or k == 'timestep' or k == 'upgrade_transformers':
                continue
            elif not Path(v).is_absolute():
                data[k] = str(Path(self.module_path) / v)
        return data

    def _get_all_voltages(self):
        """Computes over and under voltages for all buses"""
        voltage_dict = {}
        bus_names = dss.Circuit.AllBusNames()
        for b in bus_names:
            dss.Circuit.SetActiveBus(b)
            vang = dss.Bus.puVmagAngle()
            if len(vang[::2]) > 0:
                vmag = sum(vang[::2])/len(vang)
            else:
                vmag = 0
            voltage_dict[b] = vmag*2

        return voltage_dict

    def _get_line_loading(self):
        """Computes the loading for Lines."""

        line_overloads_dict = {}
        # Set the active class to be the lines
        dss.Circuit.SetActiveClass("Line")

        # Loop over the lines
        flag = dss.ActiveClass.First()
        while flag > 0:
            line_name = dss.CktElement.Name()
            # Get the current limit
            line_limit = dss.CktElement.NormalAmps()

            # Compute the current through the line
            phase = int(.5*len(dss.CktElement.Currents()))
            line_current = [math.sqrt(dss.CktElement.Currents()[2*(ii-1)+2]**2 + dss.CktElement.Currents()[2*ii+1]**2) for ii in range(phase)]

            # The loading is the ratio of the two
            ldg = max(line_current)/float(line_limit)
            line_overloads_dict[line_name] = max(line_current)/float(line_limit)

            # Move on to the next line
            flag = dss.ActiveClass.Next()
        return line_overloads_dict

    def _get_xfmr_overloads(self, ub=1.0):

        #####################################
        #    Transformer current violations
        #####################################
        #
        transformer_violation_dict = {}
        dss.Circuit.SetActiveClass("Transformer")
        flag = dss.ActiveClass.First()
        while flag > 0:
            # Get the name of the Transformer
            transformer_name = dss.CktElement.Name()
            transformer_current = []

            # transformer_limit = dss.CktElement.NormalAmps()

            hs_kv = float(dss.Properties.Value('kVs').split('[')[1].split(',')[0])
            kva = float(dss.Properties.Value('kVA'))
            n_phases = dss.CktElement.NumPhases()
            if n_phases > 1:
                transformer_limit_per_phase = kva/(hs_kv*math.sqrt(3))
            else:
                transformer_limit_per_phase = kva/hs_kv

            # nwindings = int(dss.Properties.Value("windings"))
            primary_bus = dss.Properties.Value("buses").split('[')[1].split(',')[0]

            # phase = int((len(dss.CktElement.Currents())/(nwindings*2.0)))
            Currents = dss.CktElement.CurrentsMagAng()[:2*n_phases]
            Current_magnitude = Currents[::2]

            transformer_current = Current_magnitude

            # Compute the loading
            ldg = max(transformer_current)/transformer_limit_per_phase
            transformer_violation_dict[transformer_name] = ldg

            # Move on to the next Transformer...
            flag = dss.ActiveClass.Next()

        return transformer_violation_dict

    def run(self):

        from ditto.store import Store
        from ditto.writers.opendss.write import Writer
        from urbanopt_ditto_reader.reader.read import Reader

        from ditto.consistency.check_loops import check_loops
        from ditto.consistency.check_loads_connected import check_loads_connected
        from ditto.consistency.check_unique_path import check_unique_path
        from ditto.consistency.check_matched_phases import check_matched_phases
        from ditto.consistency.check_transformer_phase_path import check_transformer_phase_path

        from ditto.consistency.fix_transformer_phase_path import fix_transformer_phase_path
        from ditto.consistency.fix_undersized_transformers import fix_undersized_transformers

        model = Store()

        reader = Reader(
            geojson_file=self.geojson_file,
            equipment_file=self.equipment_file,
            load_folder=self.urbanopt_scenario,
            use_reopt=self.use_reopt,
            is_timeseries=True,
            timeseries_location=self.timeseries_location,
            relative_timeseries_location=os.path.join('..', 'profiles')
        )
        reader.parse(model)

        OKGREEN='\033[92m'
        FAIL = '\033[91m'
        ENDC = '\033[0m'

        final_pass = True

        print('Check no loops:',flush=True)
        loops_res = check_loops(model,verbose=True)
        final_pass = final_pass and loops_res
        result = 'FAIL'
        color = FAIL
        if loops_res:
            result = 'PASS'
            color = OKGREEN
        print('Result:', f'{color} {result} {ENDC}')
        print()

        print('Check loads connected to source:',flush=True)
        result = 'FAIL'
        color = FAIL
        loads_connected_res = check_loads_connected(model,verbose=True)
        final_pass = final_pass and loads_connected_res
        if loads_connected_res:
            result = 'PASS'
            color = OKGREEN
        print('Result:', f'{color} {result} {ENDC}')
        print()

        print('Check unique path from each load to source:',flush=True)
        unique_path_res = check_unique_path(model,show_all=True,verbose=True)
        final_pass = final_pass and unique_path_res
        result = 'FAIL'
        color = FAIL
        if unique_path_res:
            result = 'PASS'
            color = OKGREEN
        print('Result:', f'{color} {result} {ENDC}')
        print()

        print('Check that phases on either side of transformer are correct:',flush=True)
        matched_phases_res = check_matched_phases(model,verbose=True)
        final_pass = final_pass and check_matched_phases
        result = 'FAIL'
        color = FAIL
        if matched_phases_res:
            result = 'PASS'
            color = OKGREEN
        print('Result:', f'{color} {result} {ENDC}')
        print()

        print('Check that phases from transformer to load and source are correct:',flush=True)
        transformer_phase_res = check_transformer_phase_path(model,needs_transformers=True, verbose=True)

        # don't do check here - see if we can fix it first
        #final_pass = final_pass and transformer_phase_res
        result = 'FAIL'
        color = FAIL
        if transformer_phase_res:
            result = 'PASS'
            color = OKGREEN

        #print('Result:', f'{color} {result} {ENDC}')
        print()

        print('Attempting to fix phases from transformer to load and source', flush=True)
        if result == 'FAIL':
            fix_transformer_phase_path(model,needs_transformers=True, verbose=True)
            transformer_phase_res = check_transformer_phase_path(model,needs_transformers=True, verbose=True)
            final_pass = final_pass and transformer_phase_res
            result = 'FAIL'
            color = FAIL
            if transformer_phase_res:
                result = 'PASS'
                color = OKGREEN
            print('Result:', f'{color} {result} {ENDC}')
            print()

        if self.upgrade_transformers:
            print('Upgrading undersized transformers:',flush=True)
            fix_undersized_transformers(model,verbose=True)


        if not final_pass:
            raise ValueError("Geojson file input structure incorrect")

        if not os.path.exists(os.path.join(self.dss_analysis, 'dss_files')):
            os.makedirs(os.path.join(self.dss_analysis, 'dss_files'), exist_ok=True)
        if not os.path.exists(os.path.join(self.dss_analysis, 'results', 'Features')):
            os.makedirs(os.path.join(self.dss_analysis, 'results', 'Features'), exist_ok=True)
        if not os.path.exists(os.path.join(self.dss_analysis, 'results', 'Transformers')):
            os.makedirs(os.path.join(self.dss_analysis, 'results', 'Transformers'), exist_ok=True)
        if not os.path.exists(os.path.join(self.dss_analysis, 'results', 'Lines')):
            os.makedirs(os.path.join(self.dss_analysis, 'results', 'Lines'), exist_ok=True)

        writer = Writer(output_path=os.path.join(self.dss_analysis, 'dss_files'), split_feeders=False, split_substations=False)
        writer.write(model)

        ts = pd.read_csv(os.path.join(self.timeseries_location, 'timestamps.csv'), header=0)
        number_iterations = len(ts)

        voltage_df_dic = {}
        line_df_dic = {}
        transformer_df_dic = {}

        delta = datetime.datetime.strptime(ts.loc[1]['Datetime'], '%Y/%m/%d %H:%M:%S') - datetime.datetime.strptime(ts.loc[0]['Datetime'], '%Y/%m/%d %H:%M:%S')
        interval = delta.seconds/3600.0
        stepsize = 60*interval
        building_map = {}

        geojson_content = []
        try:
            with open(self.geojson_file, 'r') as f:
                geojson_content = json.load(f)
        except:
            raise IOError("Problem trying to read json from file " + self.geojson_file)

        for element in geojson_content["features"]:
            if 'properties' in element and 'type' in element['properties'] and 'buildingId' in element['properties'] and element['properties']['type'] == 'ElectricalJunction':
                building_map[element['properties']['id']] = element['properties']['buildingId']

        if self.start_time is not None and self.end_time is not None:
            print(f'Attempting to run from {self.start_time} to {self.end_time}:')
        else:
            print('Running for all timepoints:')

        start_index_entry = ts.Datetime[ts.Datetime==self.start_time].index
        if len(start_index_entry) == 0:
            print(f'Warning - start time of {self.start_time} not found in timeseries file {os.path.join(self.timeseries_location, "timestamps.csv")}. Running for all times...')
            start_index = 0
        if len(start_index_entry) >1:
            print(f'Warning - start time of {self.start_time} has duplicate entries in timeseries file {os.path.join(self.timeseries_location, "timestamps.csv")}. Running for all times...')
            start_index = 0
        if len(start_index_entry) ==1:
            print(f'Unique starting time of {self.start_time} found')
            start_index = start_index_entry[0]

        end_index_entry = ts.Datetime[ts.Datetime==self.end_time].index
        if len(end_index_entry) == 0:
            print(f'Warning - end time of {self.end_time} not found in timeseries file {os.path.join(self.timeseries_location, "timestamps.csv")}. Running for all times...')
            end_index = len(ts)
        if len(end_index_entry) >1:
            print(f'Warning - end time of {self.end_time} has duplicate entries in timeseries file {os.path.join(self.timeseries_location, "timestamps.csv")}. Running for all times...')
            end_index = len(ts) 
        if len(end_index_entry) ==1:
            print(f'Unique ending time of {self.end_time} found')
            end_index = end_index_entry[0]+1

        if self.timestep is None or not (isinstance(self.timestep,float) or isinstance(self.timestep,int)):
            print(f'Using default timestep of {stepsize} minutes')
            self.timestep = stepsize
        else:
            if not self.timestep%stepsize == 0:
                raise ValueError(f"Timestep {self.timestep} is not a multiple of the loadfile step size of {stepsize}")
            print(f'Using timestep of {self.timestep} minutes')
        for i in range(start_index,end_index,int(self.timestep/stepsize)):
            time = ts.loc[i]['Datetime']
            print('Timepoint:',time,flush=True)
            hour = int(i/(1/(self.timestep/60.0)))
            seconds = (i % (1/(self.timestep/60.0)))*3600
            location = os.path.join(self.dss_analysis, 'dss_files', 'Master.dss')
            dss.run_command("Clear")
            output1 = dss.run_command("Redirect "+location)
            output2 = dss.run_command("Solve mode=yearly stepsize="+str(self.timestep)+"m number=1 hour="+str(hour)+" sec="+str(seconds))
            voltages = self._get_all_voltages()
            line_overloads = self._get_line_loading()
            overloaded_xfmrs = self._get_xfmr_overloads()

            for element in voltages:
                if element not in building_map:
                    continue
                building_id = building_map[element]
                if building_id not in voltage_df_dic:
                    voltage_df_dic[building_id] = pd.DataFrame(columns=['Datetime', 'p.u. voltage', 'overvoltage', 'undervoltage'])
                voltage_df_dic[building_id].loc[i] = [time, voltages[element], voltages[element] > 1.05, voltages[element] < 0.95]
            for element in line_overloads:
                if element not in line_df_dic:
                    line_df_dic[element] = pd.DataFrame(columns=['Datetime', 'p.u. loading', 'overloaded'])
                line_df_dic[element].loc[i] = [time, line_overloads[element], line_overloads[element] > 1.0]
            for element in overloaded_xfmrs:
                if element not in transformer_df_dic:
                    transformer_df_dic[element] = pd.DataFrame(columns=['Datetime', 'p.u. loading', 'overloaded'])
                transformer_df_dic[element].loc[i] = [time, overloaded_xfmrs[element], overloaded_xfmrs[element] > 1.0]

        for element in voltage_df_dic:
            voltage_df_dic[element].to_csv(os.path.join(self.dss_analysis, 'results', 'Features', element+'.csv'), header=True, index=False)
        for element in line_df_dic:
            line_df_dic[element].to_csv(os.path.join(self.dss_analysis, 'results', 'Lines', element+'.csv'), header=True, index=False)
        for element in transformer_df_dic:
            transformer_df_dic[element].to_csv(os.path.join(self.dss_analysis, 'results', 'Transformers', element+'.csv'), header=True, index=False)



"""
****************************************************************************************************
:copyright (c) 2019-2021 URBANopt, Alliance for Sustainable Energy, LLC, and other contributors.
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:
Redistributions of source code must retain the above copyright notice, this list of conditions
and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions
and the following disclaimer in the documentation and/or other materials provided with the
distribution.
Neither the name of the copyright holder nor the names of its contributors may be used to endorse
or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
****************************************************************************************************
"""

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="urbanopt-ditto-reader",
    version="0.3.11",
    author="Tarek Elgindy",
    author_email="tarek.elgindy@nrel.gov",
    description="Enhancement of URBANopt GeoJSON that can be consumed by DiTTo reader",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/urbanopt/urbanopt-ditto-reader",
    license="Custom",
    packages=find_packages(exclude=("tests", "docs", "example")),
    python_requires='>=3.7',
    py_modules=['ditto_reader_cli'],
    include_package_data=True,
    install_requires=[
            'ditto.py~=0.2.1', 
            'opendssdirect.py~=0.6.1', 
            'pandas~=1.2.4'
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    entry_points='''
        [console_scripts]
        ditto_reader_cli=urbanopt_ditto_reader.ditto_reader_cli:cli
        update_licenses=update_licenses:update_licenses
    '''
)
